/* This file is autogenerated by ./configure -- do not edit! */
#define FLINT_USES_POPCNT
#define FLINT_USES_BLAS 1
#define FLINT_USES_TLS 1
#define FLINT_USES_FENV 1
#define FLINT_USES_PTHREAD 1
#define FLINT_USES_GC 0
#define FLINT_USES_CPUSET 1
#define FLINT_REENTRANT 0
#define FLINT_WANT_ASSERT 0
#ifdef FLINT_USE_DLL
#define FLINT_DLL __declspec(dllimport)
#else
#define FLINT_DLL __declspec(dllexport)
#endif
